------------------------------------------------------------------------------------------
-- Contents:
--		CloseAll - Close all active BFDtools
--		ShowAll - Show currently active BFDtools
--		WipeINI - Delete default INI file
--		OpenINI - Opens the default INI file
--		CreateBFDtool - Create a BFDtool based after a skeleton template.
--
-- Requires:
--		jbFunctions.ms
--
-- History:
--		11.10.1999 - Created
--		03.18.2001 - Updated (crap... have I been using this THAT long???)
------------------------------------------------------------------------------------------

macroScript CloseAll
category:"BFDtoolsDevSupport"
buttontext:"Close All Active BFDtools"
tooltip:"Close All Active BFDtools"
(
	BFDman.closeAllTools()
)

macroScript ShowAll
category:"BFDtoolsDevSupport"
buttontext:"Show Active BFDtools"
tooltip:"Show Active BFDtools"
(
	format "\nActive Tool Count: %\n" (BFDman.numTools())

	for i in 1 to BFDman.numTools() do (
		format "\tTool %: %\n" i (BFDman.getTool i).toolName
	)

	format "\n"
)

macroScript WipeINI
category:"BFDtoolsDevSupport"
buttontext:"Delete default BFDtools INI file"
tooltip:"Delete default BFDtools INI file"
(
	try (
		print (deleteFile (GetDir #plugcfg + "\\BFDtools.ini"))
	) catch ( print false )
)

macroScript OpenINI
category:"BFDtoolsDevSupport"
buttontext:"Open default BFDtools INI file"
tooltip:"Open default BFDtools INI file"
(
	try (
		edit (GetDir #plugcfg + "\\BFDtools.ini")
	) catch ( print false )
)

macroScript CreateBFDtool
category:"BFDtoolsDevSupport"
buttontext:"Create a BFDtool skeleton"
tooltip:"Create a BFDtool skeleton"
(
	local SKELETONFILE = (getDir #maxRoot) + "\\plugcfg\\BFDtoolsDevSupport_BFDtoolSkeleton.txt"
	local AUTHORLIST = #("John Burnett","Jared Keller","Sexy Monkey","Your Name Here")
	local CATEGORYLIST = #("BFDtools","Your Catagory Here")

	local thisTool = BFDtool	toolName:"CreateBFDtool"	\
								author:"John Burnett"		\
								createDate:[11,10,1999]		\
								modifyDate:[11,10,1999]		\
								version:1					\
								defFloaterSize:[200,270]

	if maxMasterDir != undefined then (
		if ((MaxVersion())[1] >= 4000) then
			SKELETONFILE = (maxMasterDir + "\\MacroScripts\\JBurnettAF\\BFDtoolsDevSupport\\BFDtoolsDevSupport_BFDtoolSkeleton.txt")
		else
			SKELETONFILE = (maxMasterDir + "\\ScriptsUser\\JBurnettAF\\BFDtoolsDevSupport\\BFDtoolsDevSupport_BFDtoolSkeleton.txt")
	)

	rollout DLGaboutRollout "About" (
		label DLGAbout01 ""
		label DLGAbout02 ""
		label DLGAbout03 ""

		on DLGaboutRollout open do (
			DLGabout01.text = thisTool.toolName
			DLGabout02.text = thisTool.author
			DLGabout03.text =	(thisTool.modifyDate.x as integer) as string + "." +
								(thisTool.modifyDate.y as integer) as string + "." +
								(thisTool.modifyDate.z as integer) as string
		)

		on DLGaboutRollout close do ( thisTool.closeTool() )
	)

	rollout DLGmainRollout "Main" (
		fn updateUI = (
			DLGmainRollout.DLGcreateScript.enabled =	DLGmainRollout.DLGtoolName.text != "" AND
														DLGmainRollout.DLGtoolDesc.text != "" AND
														DLGmainRollout.DLGauthor.text != "" AND
														DLGmainRollout.DLGmacroCategory.text != ""
		)

		label DLGtoolNameLabel "ToolName:" align:#left
		edittext DLGtoolName "" offset:[53,-19] fieldWidth:93

		label DLGtoolDescLabel "Tool Desc:" align:#left
		edittext DLGtoolDesc "" offset:[53,-19] fieldWidth:93

		label DLGauthorLabel "Author:" align:#left
		edittext DLGauthor "" offset:[35,-19] fieldWidth:95
		dropdownlist DLGauthorList "" offset:[137,-24] items:AUTHORLIST width:14

		label DLGmacroCategoryLabel "MacroCategory:" align:#left
		edittext DLGmacroCategory "" offset:[75,-19] fieldWidth:55
		dropdownlist DLGmacroCategoryList "" offset:[137,-24] items:CATEGORYLIST width:14

		group "Auto Save/Load:" (
			checkbox DLGautoLoadRolloutStates "Rollout States" checked:true
			checkbox DLGautoLoadFloaterSize "Floater Geometry" checked:true
		)

		button DLGcreateScript "Create Script" align:#center enabled:false

		on DLGtoolName changed text do ( updateUI() )
		on DLGtoolDesc changed text do ( updateUI() )
		on DLGauthor changed text do ( updateUI() )
		on DLGmacroCategory changed text do ( updateUI() )

		on DLGauthorList selected idx do (
			DLGauthor.text = DLGauthorList.selected
			updateUI()
		)

		on DLGmacroCategoryList selected idx do (
			DLGmacroCategory.text = DLGmacroCategoryList.selected
			updateUI()
		)

		on DLGcreateScript pressed do (
			local toolName = DLGtoolName.text
			local toolDesc = DLGtoolDesc.text
			local createDate = dateAsPoint3 localTime
			local macroCategory = DLGmacroCategory.text
			local author = DLGauthor.text
			local loadRollStates = DLGautoLoadRolloutStates.checked
			local loadFloaterSize = DLGautoLoadFloaterSize.checked
			local jbVer = jbFunctionsCurrentVersion()

			local tags = #(
				#ToolName,
				#ToolDesc,
				#CreateDate,
				#MacroCategory,
				#AuthorName,
				#LoadRollStates,
				#LoadFloaterSize,
				#JBFunctionsVersion)
			local tagVals = #(
				toolName,
				toolDesc,
				createDate,
				macroCategory,
				author,
				loadRollStates,
				loadFloaterSize,
				jbVer)

			local toolStream = newScript()
			local f = openFile SKELETONFILE
			if f != undefined then (
				while NOT eof f do (
					local curLine = readLine f
					local str = replaceTags curLine tags tagVals
					format "%\n" str to:toolStream
				)
				close f
			) else (
				messageBox ("Skeleton File:\n\n" + SKELETONFILE + "\n\nNot Found!") title:"Error"
			)
		)

		on DLGmainRollout open do (
			DLGauthor.text = DLGauthorList.selected
			DLGmacroCategory.text = DLGmacroCategoryList.selected
			updateUI()
		)
	)

	thisTool.addRoll #(DLGaboutRollout, DLGmainRollout) rolledUp:#(true,false)

	thisTool.openTool thisTool
)
